#include "MediaStreamer.h"
#include <string>

static const char *subversion_id = "$Id: MediaStreamer.cpp 304 2004-03-25 19:52:41Z tkay $";

using namespace MIMD;

MediaStreamer::MediaStreamer(const Source &src) {
   theSource = new Source(src);
   scheduler = BasicTaskScheduler::createNew();
   env = BasicUsageEnvironment::createNew(*scheduler);
   audioSink = NULL;
   videoSink = NULL;
   setupDone = false;
   rtpGroupsockAudio = NULL;
   rtcpGroupsockAudio = NULL;
   rtpGroupsockVideo = NULL;
   rtcpGroupsockVideo = NULL;
}

MediaStreamer::~MediaStreamer() 
{ 
   if(theSource) delete theSource; 
   if(rtpGroupsockAudio) delete rtpGroupsockAudio;
   if(rtcpGroupsockAudio) delete rtcpGroupsockAudio;
   if(rtpGroupsockVideo) delete rtpGroupsockVideo;
   if(rtcpGroupsockVideo) delete rtcpGroupsockVideo;
}

RTPSink *MediaStreamer::getAudioSink() {
   if(!hasAudio()) return NULL;
   return audioSink;
}

RTPSink *MediaStreamer::getVideoSink() {
   if(!hasVideo()) return NULL;
   return videoSink;
}

void MediaStreamer::run() {
   // Finalize network and sink setup
   doSetup();

   // Move to the first source
   if(!moveToNextSource()) {
      cerr << "Mimd: That's odd. MediaStreamer was created without "
              "anything to play." << endl;
      return;
   }

   // Queue the source onto the sink
   queue();

   // Process the loop. This function doesn't return until we are out
   // of stuff to do (actually, pthread_exit() is called).
   // See afterPlaying() for the entry point of the event loop callbacks.
   getScheduler()->doEventLoop();
}

void MediaStreamer::queue()
{
   // Queue up the first source. Always start the video first for
   // sync purposes. These are only called once here, but the
   // event loop processing will hit them again when it calls
   // afterPlaying().
   if(hasVideo() && videoSink != NULL) {
      FramedSource *src = getVideoFSource();
      if(src == NULL) {
         cerr << "Mimd: Framed video source was"
                 " not supplied in MediaStreamer::queue()." << endl;
      } else
         videoSink->startPlaying(*src, MediaStreamer::afterPlaying, 
                                 this);
   }
      
   if(hasAudio() && audioSink != NULL) {
      FramedSource *src = getAudioFSource();
      if(src == NULL) {
         cerr << "Mimd: Framed audio source was "
                 "not supplied in MediaStreamer::queue()." << endl;
      } else
         audioSink->startPlaying(*src, MediaStreamer::afterPlaying,
                                 this);
   }
}

void MediaStreamer::afterPlaying(void *arg)
{
   MediaStreamer *player = static_cast<MediaStreamer*>(arg);

   // Make sure both halves are done.
   if ((player->hasAudio() && 
        player->getAudioFSource()->isCurrentlyAwaitingData()) || 
       (player->hasVideo() && 
        player->getVideoFSource()->isCurrentlyAwaitingData())) 
       return;

   // Make sure the currently playing stuff gets stopped.
   if(player->hasAudio())
      player->getAudioSink()->stopPlaying();

   if(player->hasVideo())
      player->getVideoSink()->stopPlaying();

   // Move to the next one. This call is required to do cleanup of the 
   // actual framed sources. This is only responsible for stopping them.
   if(!player->moveToNextSource())
      // No next source...FIXME: this may be better as return instead.
      pthread_exit(NULL); 

   // Queue up the next set.
   player->queue();

   return; // To the event loop processing
}

void MediaStreamer::doSetup()
{
   if(setupDone) return;
   setupDone = true;
   // Set up the network stuff
   const unsigned char ttl = getSource().getTTL(); 
   const unsigned maxCNAMElen = 100;
   unsigned char CNAME[maxCNAMElen+1];
   gethostname((char*)CNAME, maxCNAMElen);
   CNAME[maxCNAMElen] = '\0'; // just in case

   if(hasAudio()) {
      std::string destinationAddressStrA = getSource().getAG();
      const unsigned short rtpPortNumAudio = getSource().getAP();
      const unsigned short rtcpPortNumAudio = getSource().getAP()+1;
      struct in_addr destinationAddressA;
      destinationAddressA.s_addr = 
         inet_addr(destinationAddressStrA.c_str());
      const Port rtpPortAudio(rtpPortNumAudio);
      const Port rtcpPortAudio(rtcpPortNumAudio);
      rtpGroupsockAudio = new Groupsock(*env, destinationAddressA, 
                                  rtpPortAudio, ttl);
      rtcpGroupsockAudio = new Groupsock (*env, destinationAddressA, 
                                   rtcpPortAudio, ttl);
      // Create a 'MPEG Audio RTP' sink from the RTP 'groupsock':
      audioSink = MPEG1or2AudioRTPSink::createNew(*env, rtpGroupsockAudio);
      // Create (and start) a 'RTCP instance' for this RTP sink:
      const unsigned totalSessionBandwidthAudio = 160; // in kbps;
      RTCPInstance::createNew(*env, rtcpGroupsockAudio,
           totalSessionBandwidthAudio, CNAME,
           audioSink, NULL , false);
      // Note: This starts RTCP running automatically
   }
   if(hasVideo()) {
      std::string destinationAddressStrV = getSource().getVG();
      const unsigned short rtpPortNumVideo = getSource().getVP();
      const unsigned short rtcpPortNumVideo = getSource().getVP()+1;
      struct in_addr destinationAddressV;
      destinationAddressV.s_addr = 
         inet_addr(destinationAddressStrV.c_str());
      const Port rtpPortVideo(rtpPortNumVideo);
      const Port rtcpPortVideo(rtcpPortNumVideo);
      rtpGroupsockVideo = new Groupsock(*env, destinationAddressV, 
                                  rtpPortVideo, ttl);
      rtcpGroupsockVideo = new Groupsock(*env, destinationAddressV, 
                                   rtcpPortVideo, ttl);
      // Create a 'MPEG Video RTP' sink from the RTP 'groupsock':
      videoSink = MPEG1or2VideoRTPSink::createNew(*env, rtpGroupsockVideo);

      // Create (and start) a 'RTCP instance' for this RTP sink:
      const unsigned totalSessionBandwidthVideo = 4500; // in kbps;
      RTCPInstance::createNew(*env, rtcpGroupsockVideo,
           totalSessionBandwidthVideo, CNAME,
           videoSink, NULL /* we're a server */, false);
   }
}
