#include <string>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "MediaType.h"
#include "parser.h"

using namespace MIMD;

static const char *subversion_id = "$Id: MediaType.cpp 113 2003-07-10 18:39:47Z  $";

/**
 * Skip and ID3 headers.
 *
 * @return true if there were ID3 headers, false otherwise.
 */
static bool skip_id3(int fd)
{
   unsigned char buf[10];
   bool foundID3=false;
   int rv;
   off_t skip;

   while(1) // Skip however many there are
   {
      if((rv=read(fd, buf, 10)) < 10) { 
         // not enough data, or there was an error
         if(rv < 1) break;
         lseek(fd, -rv, SEEK_CUR); // undo the partial read
         break;
      }
      if(strncmp((char*)buf, "ID3", 3) == 0) { // Got one!
         // Skip the header. 
         skip = ((buf[6]&0x7F)<<21) + ((buf[7]&0x7F)<<14) +
                ((buf[8]&0x7F)<<7) + (buf[9]&0x7F);
         lseek(fd, skip, SEEK_CUR);
         foundID3=true;
      } else {
         lseek(fd, -10, SEEK_CUR); // No ID3
         break;
      }
   }

   return foundID3;
}

/**
 * Read the "magic" of a given file, and try to predict the media type.
 * This function is currently written for the typical big endian systems 
 * (bit order). It will not work properly on little endian systems.
 */
MediaType::MediaDescriptor MediaType::check(const char *filename)
{
   MediaDescriptor md;
   int fd;
   unsigned char sig[100];

   if((fd=open(filename, O_RDONLY)) == -1)
      return md;

   md.hasID3 = skip_id3(fd); // Skip any ID3 tags
   read(fd, sig, sizeof(sig));
   close(fd);

   // Check for MPEG
   if((sig[0] & 0xff) == 0xff && (sig[1] & 0xfe) == 0xfa) md.type = MPEG1L3;
   if((sig[0] & 0xff) == 0xff && (sig[1] & 0xfe) == 0xfc) md.type = MPEG1L2;
   if((sig[0] & 0xff) == 0xff && (sig[1] & 0xf8) == 0xf0) {
      if(sig[1] & 0x02) md.type = MPEG2L3;
      if(sig[1] & 0x04) md.type = MPEG2L2;
   }
   if((sig[0] == 0x00) && (sig[1] == 0x00) &&
      (sig[2] == 0x01) && (sig[3] == 0xb3)) md.type = MPEGVS;

   if((sig[0] == 0x00) && (sig[1] == 0x00) &&
      (sig[2] == 0x01) && (sig[3] == 0xba)) md.type = MPEGSS;

   // Check for Apple
   if(strncmp((char*)sig+4, "moov", 4) == 0) md.type = QTMOV;
   if(strncmp((char*)sig+4, "mdat", 4) == 0) md.type = QTMDAT;

   // Check for MS
   if(strncmp((char*)sig+8, "AVI", 3) == 0) md.type = AVI;
   if(strncmp((char*)sig+8, "WAVE", 4) == 0) md.type = WAV;

   // ******************** KLUDGE ****************
   if(md.hasID3 && md.type == Unknown) {
      // has an ID3 tag, but some sort of unsupported header junk.
      // Assume it is an MP3 (until I have time to find better docs on MP3 
      // header formats).
      md.type = MPEG1L3; 
   }
   return md;
}

static const char *descs[] = {
  "UNKNOWN", "MPEG 1.0 Layer 3", "MPEG 1.0 Layer 2", "MPEG 2.0 Layer 3",
  "MPEG 2.0 Layer 2", "MPEG video stream", "MPEG system stream",
  "Apple Quicktime (moov)", "Apple Quicktime (mdat)", "WAV", "AVI", 
  "No File Found" };

std::string MediaType::getDescription(MediaType::MediaID id)
{
   if(id < Unknown || id > NoFile) return std::string("BAD ID");
   return std::string(descs[id]);
}
