#ifndef MIMD_MEDIASTREAMER
#define MIMD_MEDIASTREAMER

#include "parser.h"
#include <BasicUsageEnvironment.hh>
#include <RTPSink.hh>
#include <FramedSource.hh>
#include <MPEG1or2DemuxedElementaryStream.hh>
#include <MPEG1or2Demux.hh>
#include <MPEG1or2AudioRTPSink.hh>
#include <MPEG1or2VideoRTPSink.hh>
#include <MPEG1or2VideoStreamFramer.hh>
#include <MPEG1or2AudioStreamFramer.hh>
#include <Groupsock.hh>
#include <GroupsockHelper.hh>
#include <RTCP.hh>
#include <ByteStreamFileSource.hh>
#include <MediaSource.hh>

namespace MIMD
{
   /**
    * This is the main interface to a streamer.
    */
   class MediaStreamer
   {
      public:
         /**
          * Sets up the networking and liveMedia objects based on the Source
          * object.
          */
         MediaStreamer(const Source &src);
         virtual ~MediaStreamer();

         /** 
          * @return true if there should be video
          */
         virtual bool hasVideo() = 0;

         /** 
          * @return true if there should be audio
          */
         virtual bool hasAudio() = 0;

         /**
          * Hook for a subclass to provide a FrameSource for audio. 
          * MUST be overridden if you have audio.
          *
          * @return A Framed source that is ready to be streamed via RTP.
          * Multiple calls to this method should always return the same object
          * unless a call to moveToNextSource() is made.
          */
         virtual FramedSource *getAudioFSource() { return NULL; }

         /**
          * Hook for a subclass to provide a FrameSource for audio. 
          * MUST be overridden if you have video.
          *
          * @return A Framed source that is ready to be streamed via RTP.
          * Multiple calls to this method should always return the same object
          * unless a call to moveToNextSource() is made.
          */
         virtual FramedSource *getVideoFSource() { return NULL; }

         /**
          * Called when the streamer is about to ask for new framed sources
          * from getAudioFSource()/getVideoFSource(). This is where you should
          * clean up the source that was playing, and prepare to supply a new
          * one.
          *
          * This functions IS called prior to playing the first stream, so your
          * subclass should pretend to be "one before" the first source (so
          * that moving the the next one puts you on the first one).
          * 
          * This function is responsible for implementing looping behavior. 
          * If it returns false, then the MediaStreamer in question considers
          * everything to be done, and completes it's thread of control.
          *
          * @return true if there is another source, false otherwise.
          */
         virtual bool moveToNextSource() = 0;

         /**
          * @return A properly built and connected audio sink, or NULL if this
          * streamer doesn't support audio.
          */
         RTPSink *getAudioSink();

         /**
          * @return A properly built and connected video sink, or NULL if this
          * streamer doesn't support video.
          */
         RTPSink *getVideoSink();

         /**
          * Run the streamer. This is your thread entry point. It calls the
          * liveMedia event loop, which in turn will do the callbacks to 
          * the afterPlaying() function, which will call the other hooks. 
          * It only returns when the streamer is done.
          */
         void run();

         // Base resource access
         const Source &getSource() { return *theSource; }
         BasicUsageEnvironment *getEnvironment() { return env; }
         TaskScheduler *getScheduler() { return scheduler; }
      private:
         void queue();

         /**
          * This method is called whenever a sink finishes playing. It insures
          * that both the audio and video are done, formally removes them from
          * the event queue (stopPlaying), figures out what to do in terms of
          * setting up the next one to play.
          *
          * The argument will be a pointer to the MediaStreamer instance 
          * that is being handled, and is recast to the proper type
          * using static_cast<>.
          */
         static void afterPlaying(void *arg);

         /**
          * This function sets up the network communication, and the sinks.
          */
         void doSetup();

         RTPSink *audioSink, *videoSink;
         Source *theSource;
         BasicUsageEnvironment *env;
         TaskScheduler *scheduler;
         bool setupDone;
         // These have to stay around:
         Groupsock *rtpGroupsockAudio;
         Groupsock *rtcpGroupsockAudio;
         Groupsock *rtpGroupsockVideo;
         Groupsock *rtcpGroupsockVideo;
   };
};

#endif

