#ifndef MIMD_XML
#define MIMD_XML
#include <cstdlib>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <iostream>
#include <fstream>
#include <list>
#include <map>
#include <stdexcept>
#include <xercesc/sax/HandlerBase.hpp>
#include <xercesc/parsers/SAXParser.hpp>
#include "MediaType.h"

/* $Id: parser.h 291 2004-03-25 18:50:57Z tkay $ */

XERCES_CPP_NAMESPACE_USE
using namespace std;

namespace MIMD
{
   /**
    * This class represents an individual item to stream (file or pipe). For
    * files, it will poll the content to see what it is. For a pipe, it cannot
    * yet determine anything useful, and is just a place to store the command.
    */
   class MediaItem
   {
    public:
      /**
       * Create a MediaItem for a file or command. If it is a file, then it will
       * read the "magic" from the file and determine if it is a supported MIMD
       * type. If there is a problem, this constructor throws a 
       * MediaType::MediaTypeError.
       *
       * @param file The name of the command or data file.
       * @param isAPipe A flag indicating if the "file" is a flat data file, or
       * a command that will produce streamable data.
       *
       * @throws MediaType::MediaTypeError if the file does not contain a 
       * supported media type, or if some other kind of problem exists with the
       * file.
       */
      MediaItem(const string &file, bool isAPipe) throw(MediaType::MediaTypeError);
      MediaItem(const MediaItem &item) { *this = item; }
      /**
       * Is this item a pipe command?
       */
      bool isPipe() const { return pipe; }
      /**
       * Get the file name of the media or pipe.
       */
      const string &getFileName() const { return name; }
      /**
       * Get the type ID (@see MediaType).
       */
      MediaType::MediaDescriptor getType() const { return type; }
      /**
       * Does this MediaItem have an audio portion?
       */
      bool hasAudio() const { return audio; } 
      /**
       * Does this MediaItem have an video portion?
       */
      bool hasVideo() const { return video; } 
    private:
      string name;
      bool pipe;
      MediaType::MediaDescriptor type;
      bool audio;
      bool video;
   };

   /**
    * Represents one of the sources in the configuration file.
    */
   class Source
   {
    public:
      /**
       * Construct a source object. This data mostly comes from the
       * configuration file. If you want an SDP file, then this constructor
       * will create it as it runs, and will delete the SDP file when the
       * destructor is called.  In other words, if MIMD exits cleanly (is not
       * killed), then it will clean up the SDP files as it goes.
       *
       * @param n The name of the source. Used for readability in messages.
       * @param vip The IP (group) for the video. Set to "" if no video is to
       * be sent.
       * @param vport The port for the video. Set to 0 if no video is to be
       * sent.
       * @param aip The IP (group) for the audio. Set to "" if no audio is to
       * be sent.
       * @param aport The port for the audio. Set to 0 if no audio is to be
       * sent.
       * @param ttl The TTL setting for the packets to be sent.
       * @param loopIt Do you want this source to endlessly loop?
       * @param genSDP This parameter controls the creation of an SDP file in
       *        sdp_dir. If true, then the source object will create such a file
       *        in the sdp_dir called [sdp_dir]/[name].sdp, all derived from
       *        the parameters of the constructor.
       * @param sdp_dir The directory in which to place SDP files.
       */
      Source(const string &n, const string &vip, unsigned short vport, 
             const string &aip, unsigned short aport, unsigned short ttlive,
             bool loopIt, bool genSDP = false, 
             const string &sdp_dir = string("/tmp"));
      ~Source();

      /*
       * Various parts of a source 
       */
      const string &getName() const { return name; }
      bool isAudioDesired() const { return (ap == 0 ? false:true); }
      bool isVideoDesired() const { return (vp == 0 ? false:true); }
      const string &getVG() const { return vg; }
      unsigned short getVP() const { return vp; }
      const string &getAG() const { return ag; }
      unsigned short getAP() const { return ap; }
      unsigned short getTTL() const { return ttl; }
      bool isLooping() const { return loop; }

      void addMediaItem(const MediaItem &item);
      const list<MediaItem> &getPlayList() const { return playList; }

    private:
      string name;
      string vg;
      unsigned short vp;
      string ag;
      unsigned short ap;
      unsigned short ttl;
      bool loop;
      string sdpFilename;
      list<MediaItem> playList;

      friend ostream &operator<<(ostream &out, const Source &src);
   };

   /**
    * A simple dump operator for Source objects. Really a debugging function.
    */
   inline ostream &operator<<(ostream &out, const Source &src)
   {
      list<MediaItem>::const_iterator i;

      out << src.name << endl <<
           "   vg=" << src.vg << endl << 
           "   vp=" << src.vp << endl <<
           "   ag=" << src.ag << endl << 
           "   ap=" << src.ap << endl << 
           "  ttl=" << src.ttl << endl << 
           " loop=" << src.loop << endl << 
           "  sdp=" << src.sdpFilename << endl << 
           "Play list: " << endl;
      for(i=src.playList.begin(); i!=src.playList.end(); i++)
         out << "   " << i->getFileName() << 
             (i->isPipe() ? " (pipe)":" (file)") << endl;

      return out;
   }

   /**
    * A class that will hold the configuration. It is also an XML SAX
    * Parser Handler (for callbacks), so many of the public methods are
    * called by the SAX XML parser (as "callbacks").
    */
   class MIMDConfiguration : public HandlerBase
   {
   public:
      MIMDConfiguration();
      ~MIMDConfiguration();

      /**
       * Get a list of the sources that were in the configuration. This is 
       * really the only function that most of the MIMD program will ever 
       * directly use.
       *
       * @return An STL list<Source*> that contains all of the sources.
       */
      const list<Source*> &getSources() { return sources; }

      // SAX override methods. No need to ever manually call these.
      void startElement(const XMLCh* const name, AttributeList& attributes);
      void endElement(const XMLCh* const name);
      void characters(const XMLCh* const chars, const unsigned int length) {}
      void ignorableWhitespace(const XMLCh* const chars, 
                               const unsigned int length) {}
      void resetDocument() {}

      // -----------------------------------------------------------------------
      //  Handlers for the SAX ErrorHandler interface
      // -----------------------------------------------------------------------
      void error(const SAXParseException& e);
      void fatalError(const SAXParseException& e);
      void warning(const SAXParseException& e);
      void resetErrors() { errors = false; }
      bool isGoodParse() { return !errors; }
   private:
      // Methods for internal parsing. 

      void setState(std::string ele, bool entering);
      void processAttributes(AttributeList &attrList);

      // State for internal parsing
      enum elements {mimd_config = 1, avsource, videosource, audiosource,
                     media};
      map<std::string, MIMDConfiguration::elements> eMap;
      bool in_avsource, in_videosource, in_audiosource, in_media, 
           in_mimd_config;
      Source *currentSource;
      bool errors;

      // Resulting configuration
      list<Source *> sources;
      string sdp_dir;
   };

};
#endif
