#include "parser.h"

static const char *subversion_id = "$Id: parser.cpp 112 2003-07-10 18:35:52Z  $";

namespace MIMD
{
   // Support function to simplify Xerces...convert XMLChar string to char*
   const char *decode(const XMLCh *s)
   {
      if(s == NULL) return "(null)";
      return XMLString::transcode(s);
   }

   // Check to see if we can read a named file
   bool isFileReadable(const string &file)
   {
      /* Best test: Try opening for read :) */
      int fd;
      if((fd = open(file.c_str(), O_RDONLY|O_NOCTTY)) == -1) return false;
      close(fd);
      return true;
   }

   /*
    * This is a "quick n dirty" check to see if we can write into a given
    * directory. From a security standpoint, it is slightly insecure, in that
    * if your SDP directory is writable, someone could create a non-writable
    * test file (.mimd_test....) so that this would fail; making a DoS attack
    * where mimd will not start. Technically, this DoS attack will not work on
    * most UNIX systems unless the 't' bit is set on the directory, making it
    * possible to delete only the files you own. On most systems, only /tmp has
    * this bit set.
    */
   bool isWritableDirectory(const string &file)
   {
      string tf = file + "/.mimd_test_you_can_safely_delete_this";
      int fd;

      // Check to see if we can make a scratch file in the directory
      if((fd=open(tf.c_str(), O_RDWR|O_CREAT|O_TRUNC)) == -1) 
         return false;

      // Close and delete the test file
      close(fd);
      unlink(tf.c_str());

      return true;
   }
}

using namespace MIMD;

MediaItem::MediaItem(const string &file, bool isAPipe) 
   throw(MediaType::MediaTypeError) : name(file), pipe(isAPipe)
{
   if(isAPipe) { // Pipes cannot be determined here yet.
      video = true; // Assume both are there
      audio = true;
      type.type = MediaType::Unknown;
      return;
   }
   audio = false;
   video = false;

   type = MediaType::check(file.c_str());
   switch(type.type)
   {
      case MediaType::MPEG1L3: 
      case MediaType::MPEG2L3:
      case MediaType::MPEG1L2:
      case MediaType::MPEG2L2:
         audio = true;
         break;
      case MediaType::MPEGSS:
         audio = true; 
         // No break...we also have video
      case MediaType::MPEGVS:
         video = true;
         break;
      case MediaType::Unknown:
         throw MediaType::MediaTypeError(file + 
            string(" - file format was not recognized"));
      case MediaType::NoFile:
         throw MediaType::MediaTypeError(file + 
            string(" - file could not be opened"));
      default:
         throw MediaType::MediaTypeError(file + string(" (") + 
            MediaType::getDescription(type.type) + 
            string(") is not supported by this version of MIMD"));
   }
}

void Source::addMediaItem(const MediaItem &item)
{ 
   playList.push_back(item); 
}

Source::~Source()
{
   // Remove the SDP file
   if(sdpFilename.length() > 0)
      unlink(sdpFilename.c_str());
}

Source::Source(const string &n, const string &vip, unsigned short vport,
               const string &aip, unsigned short aport, unsigned short ttlive,
               bool loopIt, bool genSDP, const string &sdp_dir
              ) : name(n), vg(vip), vp(vport), ag(aip), ap(aport), ttl(ttlive), 
                  loop(loopIt)
{
   char hostname[50];
   char value[160];

   if(!genSDP) return;

	struct timeval tv;
	unsigned long stamp;
	
	gettimeofday(&tv, 0);
	stamp = (tv.tv_sec << 16) | (tv.tv_usec & 0x0000ffff);

   sdpFilename = sdp_dir + "/" + name + ".sdp"; 
   ofstream out(sdpFilename.c_str());

	gethostname(hostname, 50);
	sprintf(value, "%s %lu %lu IN IP4 %s", getlogin(), stamp, 
                                        stamp, hostname);
   // Generate an SDP file for this source.
   out << "v=0" << endl <<
          "o=" << value << endl <<
	       "s=mimd session" << endl <<
	       "i=mimd session description" << endl <<
	       "t=0 0" << endl <<
	       "a=tool:mimd 0.2" << endl <<
	       "a=recvonly" << endl <<
	       "a=type:broadcast" << endl;
	if (ap) {
      out << "m=audio " << ap << " RTP/AVP 14" << endl <<
             "c=IN IP4 " << ag << "/" << ttl << endl;
	}
	if (vp) {
      out << "m=video " << vp << " RTP/AVP 32" << endl <<
             "c=IN IP4 " << vg << "/" << ttl << endl;
	}

   out.close();
}

/**
 * Creates the MIMDConfiguration object. This method creates an internal map
 * for element lookups, and initializes the parser's state.
 */
MIMDConfiguration::MIMDConfiguration()
{
   eMap["mimd_config"] = mimd_config;
   eMap["avsource"] = avsource;
   eMap["videosource"] = videosource;
   eMap["audiosource"] = audiosource;
   eMap["media"] = media;

   in_avsource = false;
   in_videosource = false;
   in_audiosource = false;
   in_media = false;
   currentSource = (Source*)NULL;
}

MIMDConfiguration::~MIMDConfiguration()
{
   list<Source*>::iterator i;
   for(i=sources.begin(); i!=sources.end(); i++)
      delete *i;
}

/**
 * Use a string-keyed lookup map to figure out which kind of element
 * we are seeing, and set the state of our parsing code so that it will know
 * if we are entering or leaving the block associated with that element.
 */
void MIMDConfiguration::setState(std::string ele, bool entering)
{
   switch(eMap[ele]) {
      case mimd_config:
         in_mimd_config = entering;
         break;
      case avsource:
         in_avsource = entering;
         if(!entering) cerr << *currentSource;
         break;
      case videosource:
         in_videosource = entering;
         if(!entering) cerr << *currentSource;
         break;
      case audiosource:
         in_audiosource = entering;
         if(!entering) cerr << *currentSource;
         break;
      case media:
         in_media = entering;
         break;
   }
}

/**
 * Called by the XML parser every time an end element is reached.
 */
void MIMDConfiguration::endElement(const XMLCh* const name)
{
   string elename(decode(name));

   setState(elename, false);
}

/**
 * This is the main processing for our configuration. The automatic calls
 * to startElement() and endElement() will have set the current state of
 * parsing, so we can just pull out the attributes and put them into our
 * internal state. This routine relies heavily on the XML validation system for
 * verification of a proper configuration file. There are a few sanity checks
 * here, but nothing major.
 */
void MIMDConfiguration::processAttributes(AttributeList &attrList)
{
   if(in_media) {
      if(!in_avsource && !in_videosource && !in_audiosource)
         throw runtime_error("mimd: Invalid XML configuration file.\n");
      // The attributes describe a source of multimedia data
      string type(decode(attrList.getValue("type")));
      string value(decode(attrList.getValue("value")));

      if(type == "file") {
         if(!isFileReadable(value)) {
            cerr << "Mimd: WARNING: " << value << 
                  " was not found or is not readable. SKIPPING." << endl;
            return;
         }
      }
      try {
         MediaItem item(value, (type == "file" ? false:true));
         if((in_avsource || in_videosource) && !item.hasVideo())
            cerr << "Mimd: WARNING: Failed to sense a video component in " << 
                 value << endl;
         else if((in_avsource || in_audiosource) && !item.hasAudio()) 
            cerr << "Mimd: WARNING: Failed to sense a audio component in " << 
                 value << endl;
         else
            currentSource->addMediaItem(item);
      } catch(MediaType::MediaTypeError &error) {
         cerr << "Mimd: WARNING: " << error.what() << ". Skipping." << endl;
      }
   } else if(in_avsource || in_videosource || in_audiosource) {
      // We are not in a media element, so we must be reading the attrs of vsrc
      string vs_name(decode(attrList.getValue("name")));
      string vs_vg(decode(attrList.getValue("vg")));
      string vs_ag(decode(attrList.getValue("ag")));
      unsigned short vs_vp = 
         atoi((decode(attrList.getValue("vp"))));
      unsigned short vs_ap =
         atoi(decode(attrList.getValue("ap")));
      unsigned short vs_ttl = 
         atoi(decode(attrList.getValue("ttl")));
      string vs_loop(decode(attrList.getValue("loop")));
      string generate_sdp(decode(attrList.getValue("generate_sdp")));

      currentSource = new Source(vs_name, vs_vg, vs_vp, vs_ag, vs_ap,
         vs_ttl, (vs_loop == "yes" ? true:false), 
         (generate_sdp == "yes" ? true:false), sdp_dir);
      sources.push_back(currentSource);
   } else if(in_mimd_config) {
      if(decode(attrList.getValue("sdp_dir")))
         sdp_dir = decode(attrList.getValue("sdp_dir"));
      if(!isWritableDirectory(sdp_dir)) {
         cerr << "MIMD: SDP Files cannot be stored in " << sdp_dir << ". " << 
            "Access was denied by the system." << endl;
         errors=true;
      }
   }
}

/**
 * This method is called by the XML parser every time a start element is
 * reached. Our implementation uses support functions to process the element.
 */
void MIMDConfiguration::startElement(const XMLCh* const name, 
                               AttributeList& attributes)
{
   string elename(decode(name));

   setState(elename, true);
   processAttributes(attributes);
}

void MIMDConfiguration::error(const SAXParseException& e)
{
   cerr << "\nError at file " << decode(e.getSystemId())
      << ", line " << e.getLineNumber()
      << ", char " << e.getColumnNumber()
      << "\n  Message: " << decode(e.getMessage()) << endl;
   errors=true;
}

void MIMDConfiguration::fatalError(const SAXParseException& e)
{
   cerr << "\nFatal Error at file " << decode(e.getSystemId())
      << ", line " << e.getLineNumber()
      << ", char " << e.getColumnNumber()
      << "\n  Message: " << decode(e.getMessage()) << endl;
   errors=true;
}

void MIMDConfiguration::warning(const SAXParseException& e)
{
   cerr << "\nWarning at file " << decode(e.getSystemId())
      << ", line " << e.getLineNumber()
      << ", char " << e.getColumnNumber()
      << "\n  Message: " << decode(e.getMessage()) << endl;
}
